from subprocess import PIPE, Popen
from threading import Thread
from queue import Queue
from shlex import quote

import io
import time
import re
import py_cui

from recoverpy import window_handler as WINDOW_HANDLER
from recoverpy import saver as SAVER
from recoverpy import logger as LOGGER
from recoverpy import menu_with_block_display as BLOCK_DISPLAY_MENU


class SearchMenu(BLOCK_DISPLAY_MENU.MenuWithBlockDisplay):
    """Search menu is displayed after Parameters menu.
    On the left hand scroll menu, results from the grep command will be listed.
    On the right hand textbox, result of a dd command will be displayed when the user
    selects a block.

    Args:
        BLOCK_DISPLAY_MENU (MenuWithBlockDisplay: Composition to inherit block display
        methods.

    Attributes:
        master (py_cui.PyCUI): PyCUI constructor.
        queue_object (Queue): Queue object where grep command stdout will be stored.
        result_index (int): Number of results already processed.
        partition (str): System partition selected by user for search.
        block_size (int): Size of partition block for dd parsing.
        searched_string (str): String given by the user that will be searched by dd command.
    """

    def __init__(self, master: py_cui.PyCUI, partition: str, string_to_search: str):
        """Constructor for Search menu

        Args:
            master (py_cui.PyCUI): PyCUI constructor
            partition (str): System partition to search
        """

        super().__init__()

        self.master = master

        self.queue_object = Queue()
        self.result_index = 0

        self.partition = partition
        self.block_size = 512

        self.searched_string = string_to_search

        LOGGER.write("info", "Starting 'SearchMenu' CUI window")

        self.create_ui_content()

        self.start_search()
        LOGGER.write("info", dir(self))

        LOGGER.write(
            "info",
            "Raw searched string:\n{searched_string}".format(searched_string=self.searched_string),
        )
        LOGGER.write(
            "info",
            "Formated searched string:\n{escaped_string}".format(escaped_string=quote(self.searched_string)),
        )

    def create_ui_content(self):
        """Handles the creation of the UI elements."""

        self.search_results_cell = self.master.add_scroll_menu(
            "Search results:", 0, 0, row_span=10, column_span=5, padx=1, pady=0
        )
        self.search_results_cell.add_key_command(py_cui.keys.KEY_ENTER, self.display_selected_block)

        self.result_content_box = self.master.add_text_block(
            "Block content:", 0, 5, row_span=9, column_span=5, padx=1, pady=0
        )
        self.result_content_box.add_key_command(py_cui.keys.KEY_F5, self.open_save_menu)
        self.result_content_box.add_key_command(py_cui.keys.KEY_F6, self.display_previous_block)
        self.result_content_box.add_key_command(py_cui.keys.KEY_F7, self.display_next_block)

        self.previous_button = self.master.add_button(
            "<",
            9,
            5,
            row_span=1,
            column_span=1,
            padx=1,
            pady=0,
            command=self.display_previous_block,
        )

        self.next_button = self.master.add_button(
            ">",
            9,
            8,
            row_span=1,
            column_span=1,
            padx=1,
            pady=0,
            command=self.display_next_block,
        )

        self.save_file_button = self.master.add_button(
            "Save Block",
            9,
            6,
            row_span=1,
            column_span=2,
            padx=1,
            pady=0,
            command=self.open_save_menu,
        )

        self.exit_button = self.master.add_button(
            "Exit",
            9,
            9,
            row_span=1,
            column_span=1,
            padx=1,
            pady=0,
            command=self.master.stop,
        )

    def start_search(self):
        """Function is called within __init__
        Launches a process executing the grep command.
        A thread to store the output in a queue object.
        And a second thread to populate the result box dynamically.
        """

        grep_process = Popen(
            ["grep", "-a", "-b", self.searched_string, self.partition],
            stdin=None,
            stdout=PIPE,
            stderr=None,
        )

        enqueue_grep_output_thread = Thread(
            target=self.enqueue_grep_output,
            args=(grep_process.stdout, self.queue_object),
        )
        enqueue_grep_output_thread.daemon = True
        enqueue_grep_output_thread.start()

        LOGGER.write("info", "Starting searching process")

        yield_results_thread = Thread(target=self.populate_result_list)
        yield_results_thread.daemon = True
        yield_results_thread.start()

        LOGGER.write("info", "Starting output fetching process")

    def enqueue_grep_output(self, out: io.BufferedReader, queue: Queue):
        """Function called in a thread to store the grep command output in
        a queue object.

        Args:
            out (io.BufferedReader): Output of grep process
            queue (Queue): Queue object to store stdout
        """

        for line in iter(out.readline, b""):
            queue.put(line)
        out.close()

    def yield_new_results(self):
        """Probes the queue object for new results.
        If any, returns it to populate the result box.

        Returns:
            List[str]: List of new results
        """

        new_results = []

        # Returns if no new result
        if len(list(self.queue_object.queue)) == self.result_index:
            return []

        queue_list = list(self.queue_object.queue)

        for queue_element in queue_list[self.result_index :]:
            new_results.append(queue_element)

        self.result_index = len(queue_list)

        return new_results

    def populate_result_list(self):
        """Dynamically populates the left hand result list with new results from the
        grep command"""

        while True:
            new_results = self.yield_new_results()
            for result in new_results:
                string_result = str(result)[2:-1]

                LOGGER.write("debug", "New result found: " + string_result[:30] + " ...")

                self.search_results_cell.add_item(string_result)
            self.master.set_title("{num_of_results} results".format(num_of_results=str(self.result_index)))

            # Sleeps to avoid unnecessary overload
            # This should not affect Popen's GIL
            time.sleep(2)

    def update_block_number(self):
        """Updates currently viewed block number when the user selects one in the list."""

        item_index = self.search_results_cell.get_selected_item_index()
        item_list = self.search_results_cell.get_item_list()

        item = item_list[item_index]
        inode = int(re.findall(r"^([0-9]+)\:", item)[0])
        self.current_block = str(int(inode / self.block_size))

        LOGGER.write(
            "debug",
            "Displayed block set to {current_block}".format(current_block=str(self.current_block)),
        )

    def display_selected_block(self):
        """Function called when the user select a result in the left hand list.
        Bundles updating + displaying block.
        """

        self.update_block_number()
        self.display_block(self.current_block)

    def open_save_menu(self):
        """Opens a menu displaying save options."""

        menu_choices = [
            "Save currently displayed block",
            "Explore neighboring blocks and save it all",
            "Cancel",
        ]
        self.master.show_menu_popup("How do you want to save it ?", menu_choices, self.handle_save_menu_choice)

    def handle_save_menu_choice(self, choice: str):
        """Depending on user choice, function will either directly save the output in
        a text file, open a more detailed menu called BlockMenu or just exit.

        Args:
            choice (str): User choice given by open_save_menu function.
        """
        if choice == "Save currently displayed block":
            SAVER.save_result(current_block=self.current_block, result=self.current_result)

            self.master.show_message_popup("", "Result saved.")
        elif choice == "Explore neighboring blocks and save it all":
            WINDOW_HANDLER.open_block_menu(partition=self.partition, block=self.current_block)
        elif choice == "Cancel":
            pass



from subprocess import check_output
from os import geteuid

import re
import py_cui

from recoverpy import window_handler as WINDOW_HANDLER
from recoverpy import logger as LOGGER


class ParametersMenu:
    """Parameters menu is the first window displayed.
    User is prompted to select a partition and a string to search in it.

    Attributes:
        partition_dict (dict): Dictionnary of system partitions found with
            lsblk command and their attributes.
        partition_to_search (str): Partition selected by user
        string_to_search (str): String entered by user
    """

    def __init__(self, master: py_cui.PyCUI):
        """Constructor for Parameters menu

        Args:
            master (py_cui.PyCUI): PyCUI constructor
        """

        self.master = master

        self.partition_dict = None
        self.partition_to_search = None
        self.string_to_search = None

        LOGGER.write("info", "Starting 'ParametersMenu' CUI window")

        self.is_user_root()

        self.create_ui_content()

    def create_ui_content(self):
        """Handles the creation of the UI elements."""

        self.partition_list_cell = self.master.add_scroll_menu(
            "Select a partition to search:", 0, 0, row_span=9, column_span=5
        )
        self.partition_list_cell.add_key_command(py_cui.keys.KEY_ENTER, self.select_partition)

        self.add_partitions_to_list()

        # Color rules
        self.partition_list_cell.add_text_color_rule("Mounted at", py_cui.YELLOW_ON_BLACK, "contains")
        self.partition_list_cell.set_selected_color(py_cui.GREEN_ON_BLACK)

        self.string_text_box = self.master.add_text_block("Enter a text to search:", 0, 5, row_span=9, column_span=5)

        self.start_search_button = self.master.add_button(
            "Start search",
            9,
            4,
            row_span=1,
            column_span=2,
            padx=0,
            pady=0,
            command=self.start_search,
        )

    def is_user_root(self) -> bool:
        """Checks if user has root privileges.
        The method is simply verifying if EUID == 0.
        It may be problematic in some edge cases. (Some particular OS)
        But, as grep search will not exit quickly, exception handling
        can't be used.

        Returns:
            bool: User is root
        """

        if geteuid() == 0:
            LOGGER.write("info", "User is root")
            return True

        self.master.show_error_popup("Not root", "You have to be root or use sudo.")
        LOGGER.write("warning", "User is not root")
        return False

    def list_partitions(self) -> dict:
        """Uses lsblk command to find partitions.

        Returns:
            dict: Found partitions with format :
                    {Name: FSTYPE, IS_MOUNTED, MOUNT_POINT}
        """

        partition_list = self.lsblk()

        # Create dict with relevant infos
        partition_dict = {}
        for partition in partition_list:
            if partition[2] == "":
                # Ignore if no FSTYPE detected
                continue

            if partition[3] == "":
                is_mounted = False
                mount_point = None
            else:
                is_mounted = True
                mount_point = partition[3]

            partition_dict[partition[0]] = {
                "FSTYPE": partition[2],
                "IS_MOUNTED": is_mounted,
                "MOUNT_POINT": mount_point,
            }

        # Warn the user if no partition found with lsblk
        if len(partition_dict) == 0:
            LOGGER.write("Error", "No partition found !")
            self.master.show_error_popup("Error", "No partition found.")
            return None

        LOGGER.write("debug", "Partition list generated using 'lsblk'")
        LOGGER.write(
            "debug",
            "{dict_len} partitions found".format(dict_len=str(len(partition_dict))),
        )

        return partition_dict

    def lsblk(self) -> list:
        """Uses 'lsblk' utility to generate a list of detected
        system partions."

        Returns:
            list: List of system partitions.
        """

        lsblk_output = check_output(["lsblk", "-r", "-n", "-o", "NAME,TYPE,FSTYPE,MOUNTPOINT"], encoding="utf-8")
        partition_list_raw = [line for line in lsblk_output.splitlines() if "part" in line]
        partition_list_formatted = [line.split(" ") for line in partition_list_raw]

        return partition_list_formatted

    def add_partitions_to_list(self):
        """Populates the partition list with partition found previously."""

        self.partition_dict = self.list_partitions()

        if self.partition_dict is None:
            return

        for partition in self.partition_dict:
            if self.partition_dict[partition]["IS_MOUNTED"]:
                self.partition_list_cell.add_item(
                    "Name: {name}  -  Type: {fstype}  -  Mounted at: {mountpoint}".format(
                        name=partition,
                        fstype=self.partition_dict[partition]["FSTYPE"],
                        mountpoint=self.partition_dict[partition]["MOUNT_POINT"],
                    )
                )
            else:
                self.partition_list_cell.add_item(
                    "Name: {name}  -  Type: {fstype}".format(
                        name=partition, fstype=self.partition_dict[partition]["FSTYPE"]
                    )
                )

            LOGGER.write(
                "debug",
                "Partition added to list: {partition}".format(partition=str(partition)),
            )

    def select_partition(self):
        """Handles the user selection of a partition in the list."""

        selected_partition = re.findall(r"Name\:\ ([A-Za-z0-9]+)\ ", self.partition_list_cell.get())[0]

        if self.partition_dict[selected_partition]["IS_MOUNTED"]:
            # Warn the user to unmount his partition first
            self.master.show_warning_popup(
                "Warning",
                "It is highly recommended to unmount {name} first.".format(name=selected_partition),
            )
        else:
            self.master.show_message_popup("", "Partition {name} selected.".format(name=selected_partition))

        self.partition_to_search = "/dev/" + selected_partition.strip()

        LOGGER.write(
            "info",
            "Partition selected: {partition}".format(partition=self.partition_to_search),
        )

    def start_search(self):
        """Checks if partition is selected and string is given.
        If all required elements are present, launch confirm_search function.
        """

        if not self.is_user_root():
            return

        self.string_to_search = self.string_text_box.get()

        LOGGER.write("info", "Starting search")

        if self.partition_to_search == "":
            # No partition selected
            self.master.show_message_popup("Error", "You have to select a partition to search.")
            LOGGER.write("warning", "No partition selected for search")
        elif self.string_to_search.replace(" ", "").replace("\n", "").replace("\t", "") == "":
            # Blank string to search
            self.master.show_message_popup("Error", "You have to enter a text to search.")
            LOGGER.write("warning", "No string given for search")
        else:
            # Prompt to confirm string
            self.master.show_yes_no_popup(
                "Do you want to start searching this text on partition {partition} ?".format(
                    partition=self.partition_to_search
                ),
                self.confirm_search,
            )

    def confirm_search(self, is_confirmed: bool):
        """Closes parameters menu and open search menu if confirmed.

        Args:
            is_confirmed (bool): User popup selection
        """
        if is_confirmed:
            WINDOW_HANDLER.close_parameters_menu()
            WINDOW_HANDLER.open_search_menu(
                partition=self.partition_to_search,
                string_to_search=self.string_to_search.strip(),
            )
        else:
            pass
